<?php

namespace Drupal\oidc\EventSubscriber;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\oidc\OpenidConnectSessionInterface;
use Drupal\oidc\Routing\ImmutableTrustedRedirectResponse;
use Drupal\Core\Url;
use Symfony\Component\HttpKernel\Event\FilterResponseEvent;
use Symfony\Component\HttpKernel\Event\GetResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\Routing\Exception\ResourceNotFoundException;
use Symfony\Component\Routing\Matcher\RequestMatcherInterface;

/**
 * Event subscriber for kernel events.
 */
class KernelEventSubscriber implements EventSubscriberInterface {

  /**
   * General configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $generalConfig;

  /**
   * The route matcher service.
   *
   * @var \Symfony\Component\Routing\Matcher\RequestMatcherInterface
   */
  protected $routeMatcher;

  /**
   * The OpenID Connect session service.
   *
   * @var \Drupal\oidc\OpenidConnectSessionInterface
   */
  protected $session;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Class constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory service.
   * @param \Symfony\Component\Routing\Matcher\RequestMatcherInterface $route_matcher
   *   The route matcher service.
   * @param \Drupal\oidc\OpenidConnectSessionInterface $session
   *   The OpenID Connect session service.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   */
  public function __construct(ConfigFactoryInterface $config_factory, RequestMatcherInterface $route_matcher, OpenidConnectSessionInterface $session, AccountInterface $current_user) {
    $this->generalConfig = $config_factory->get('oidc.general');
    $this->routeMatcher = $route_matcher;
    $this->session = $session;
    $this->currentUser = $current_user;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[KernelEvents::REQUEST][] = ['onRequest', 100];
    $events[KernelEvents::RESPONSE][] = 'onResponse';

    return $events;
  }

  /**
   * Alter the request before it's processed.
   *
   * @param \Symfony\Component\HttpKernel\Event\GetResponseEvent $event
   *   The kernel request event.
   */
  public function onRequest(GetResponseEvent $event) {
    try {
      $request = $event->getRequest();
      $match = $this->routeMatcher->matchRequest($request);

      // Clear the state if the request isn't to an OpenID Connect route.
      if (strpos($match['_route'], 'oidc.openid_connect.') !== 0) {
        $this->session->clearState();
      }

      // Get the redirect URL.
      $url = NULL;
      switch ($match['_route']) {
        case 'user.login':
          if ($this->currentUser->isAnonymous() && !$request->query->has('local')) {
            $url = $this->getLoginUrl($request->query->get('destination'));
          }
          break;

        case 'user.logout':
          if ($this->session->isAuthenticated()) {
            $url = $this->getLogoutUrl($request->query->get('destination'));
          }
          break;
      }

      // Set the response.
      if ($url && (!$url->isRouted() || $url->getRouteName() !== $match['_route'])) {
        $event->setResponse(new ImmutableTrustedRedirectResponse($url));
      }
    }
    catch (ResourceNotFoundException $ex) {
      // Nothing to do.
    }
  }

  /**
   * Redirect 403 responses to the login page.
   *
   * @param \Symfony\Component\HttpKernel\Event\FilterResponseEvent $event
   *   The kernel response event.
   */
  public function onResponse(FilterResponseEvent $event) {
    // Ignore if authenticated.
    if ($this->currentUser->isAuthenticated()) {
      return;
    }

    // Only respond on a 403 response.
    if ($event->getResponse()->getStatusCode() !== 403) {
      return;
    }

    // Ignore if the redirect isn't enabled.
    if (!$this->generalConfig->get('redirect_403')) {
      return;
    }

    // Redirect.
    $url = $this->getLoginUrl($event->getRequest()->getPathInfo());
    $event->setResponse(new ImmutableTrustedRedirectResponse($url));
  }

  /**
   * Get the login URL.
   *
   * @param string|null $destination
   *   The destination path.
   *
   * @return \Drupal\Core\Url
   *   URL of the (configured) login page.
   */
  protected function getLoginUrl($destination = NULL) {
    // Get the login path.
    $login_path = $this->generalConfig->get('login_path');

    // Get the options.
    $options = [];
    if ($destination !== NULL && $destination !== '') {
      $options['query']['destination'] = $destination;
    }

    if ($login_path !== NULL) {
      return Url::fromUserInput($login_path, $options);
    }

    return Url::fromRoute('user.login', [], $options);
  }

  /**
   * Get the logout URL.
   *
   * @param string|null $destination
   *   The destination path.
   *
   * @return \Drupal\Core\Url
   *   URL of the OpenID Connect logout page.
   */
  protected function getLogoutUrl($destination = NULL) {
    // Get the options.
    $options = [];
    if ($destination !== NULL && $destination !== '') {
      $options['query']['destination'] = $destination;
    }

    return Url::fromRoute('oidc.openid_connect.logout', [], $options);
  }

}
